/**
 * Copyright (C) 2007-2012 Linutop SAS <http://www.linutop.com/>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <unistd.h>
#include <sys/types.h>
#include <stdio.h>

#include <glib/gprintf.h>
#include <glib/gi18n-lib.h>

#include "liblusm.h"

gchar *device_path = NULL;
gboolean yes = FALSE;
gboolean nocheck = FALSE;
gboolean backup;
gboolean do_sync;

static GOptionEntry options[] =
{
    { "device", 'd', 0, G_OPTION_ARG_STRING, &device_path,
      N_("Install the system on DEVICE"), "DEVICE" },
    { "assume-yes", 'y', 0, G_OPTION_ARG_NONE, &yes,
      N_("Answer yes to all questions"), NULL },
    { "no-check", 'c', 0, G_OPTION_ARG_NONE, &nocheck,
      N_("Don't perform the md5 verifications"), NULL },
    { "backup", 'b', 0, G_OPTION_ARG_NONE, &backup,
      N_("Also copy the user data"), NULL },
    { "sync", 's', 0, G_OPTION_ARG_NONE, &do_sync,
      N_("Synchronise the system with the latest update"), NULL },
    { NULL }
};

static void
show_list (gpointer data, gpointer user_data)
{
    static int i = 0;
    Device *device = (Device *) data;
    gchar *size;

    size = g_format_size ((guint64) device->size);
    g_printf ("  %d: %s %s (%s)\n", ++i, device->vendor, device->model, size);
    g_free (size);
}

Device *
ask_for_device ()
{
    GList *devices = NULL;
    gint answer = 0;
    gint ret = 0;

    devices = lusm_available_targets ();
    g_list_foreach (devices, show_list, NULL);

    while ((ret < 1)
           && ((answer < 1) || (answer > g_list_length (devices))))
    {
        g_printf (_("Your choice: "));
        fflush (NULL);
        ret = scanf ("%d", &answer);
    }

    return g_list_nth_data (devices, answer-1);
}

int
main (int argc, gchar **argv)
{
    GOptionContext *context;
    Device *device = NULL;
    GError *error = NULL;
    gchar confirm[24];
    gint ret = 0;

    context = g_option_context_new (_("- Duplicate a Linutop OS"));
    g_option_context_add_main_entries (context, options, GETTEXT_PACKAGE);
    if (!g_option_context_parse (context, &argc, &argv, &error))
    {
        g_print (_("Failed to parse options: %s\n"), error->message);
        g_error_free (error);
        error = NULL;

        return 1;
    }

    if (getuid () != 0)
    {
        g_fprintf (stderr, _("You need to be root.\n"));
        return 1;
    }

    /* init lusm */
    if (!lusm_init ())
    {
        g_fprintf (stderr, _("Unable to init the lusm library."));
        return 1;
    }

    if (device_path)
    {
        if ((device = lusm_get_device_for_path (device_path)) == NULL)
        {
            g_fprintf (stderr, _("The specified device doesn't seem to be valid.\n\n"));
        }
    }

    if (!device)
    {
        device = ask_for_device ();
    }

    if (! yes)
    {
        g_printf (_("The data on device %s %s WILL BE ERASED.\n"),
                    device->vendor, device->model);
        /* TRANSLATORS: keep 'yes' untranslated; thanks. */
        g_printf (_("Type 'yes' if you wish to continue: "));
        fflush (NULL);
        ret = scanf ("%s", confirm);
        if ((ret > 0) && (strcmp (confirm, "yes") != 0))
        {
            /* TODO: free devices list */
            lusm_shutdown ();
            return 0;
        }
    }

    g_printf ("%s ", _(lusm_status_desc[LUSM_STATUS_PREP].msg));
    fflush (NULL);
    if (!lusm_partition (device, FALSE))
    {
        g_printf (_("Error\n"));
        g_fprintf (stderr, _("Impossible to setup the device. "
                             "Make sure to umount all partitions for this device.\n"));
        return 1;
    }
    else
        g_printf (_("OK\n"));

    g_printf ("%s ", _(lusm_status_desc[LUSM_STATUS_COPY].msg));
    fflush (NULL);
    if (!lusm_duplicate (NULL))
    {
        g_printf (_("Error\n"));
        g_fprintf (stderr, _("Impossible to duplicate the system.\n"));
        return 1;
    }
    else
        g_printf (_("OK\n"));

    if (!nocheck)
    {
        g_printf ("%s ", _(lusm_status_desc[LUSM_STATUS_VERIF].msg));
        fflush (NULL);
        if (!lusm_check_md5_from_file ("/cdrom/configs/files.list"))
        {
            g_printf (_("Error\n"));
            g_fprintf (stderr, _("Data verification failed. "
                                 "Continuing anyway.\n"));
        }
        else
            g_printf (_("OK\n"));
    }

    if (backup)
    {
        g_printf ("%s ", _(lusm_status_desc[LUSM_STATUS_USER].msg));
        fflush (NULL);
        if (!lusm_copy_data (NULL))
        {
            g_printf (_("Error\n"));
            g_fprintf (stderr, _("User data have not been correctly copied. "
                                 "Continuing anyway.\n"));
        }
        else
            g_printf (_("OK\n"));
    }

    g_printf ("%s ", _(lusm_status_desc[LUSM_STATUS_CONFIG].msg));
    fflush (NULL);
    if (!lusm_setup (backup))
    {
        g_printf (_("Error\n"));
        g_fprintf (stderr, _("Failed to setup the new system. "
                             "Continuing anyway.\n"));
    }
    else
        g_printf (_("OK\n"));

    g_printf ("%s ", _(lusm_status_desc[LUSM_STATUS_BOOT].msg));
    fflush (NULL);
    lusm_set_bootable (device);
    g_printf ("Ok\n");

    if (do_sync)
    {
        g_printf ("%s ", _(lusm_status_desc[LUSM_STATUS_SYNC].msg));
        fflush (NULL);
        if (!lusm_rsync ())
        {
            g_printf (_("Error\n"));
            g_fprintf (stderr, _("Failed to synchronize the system. "
                                 "Continuing anyway.\n"));
        }
        else
            g_printf (_("OK\n"));
    }

    g_printf ("Done.\n");

    lusm_shutdown ();

    return 0;
}
