/**
 * Copyright (C) 2007-2012 Linutop SAS <http://www.linutop.com/>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/wait.h>

#include "main.h"
#include "syscfg.h"

enum init_ids {
    INIT_PERSISTENT = 0,
    INIT_LOCK,
    INIT_LOCALE
};

/*
 * generic functions
 */
inline static void
set (GObject *obj, gpointer value)
{
    g_object_set_data (obj, "ltopconf", value);
}

inline static gpointer
get (GObject *obj)
{
    return g_object_get_data (obj, "ltopconf");
}

/*
 * per setting functions
 */
/* language */
void
set_locale (const gchar *code)
{
    set (G_OBJECT (dialog->locale_btn), (gpointer) g_strdup (code));
    gtk_button_set_label (GTK_BUTTON (dialog->locale_btn),
        get_locale_name (code));
}

const gchar *
get_locale_value ()
{
    return  (const gchar *) get (G_OBJECT (dialog->locale_btn));
}

/* kbd */
inline void
set_kbd_user_defined (gboolean b)
{
    g_object_set_data (G_OBJECT (dialog->kbd_btn), "user_defined",
        (gpointer) b);
}

inline gboolean
get_kbd_user_defined ()
{
    return (gboolean) g_object_get_data (G_OBJECT (dialog->kbd_btn), "user_defined");
}

void
set_kbd (const gchar *code)
{
    set (G_OBJECT (dialog->kbd_btn), (gpointer) g_strdup (code));
    gtk_button_set_label (GTK_BUTTON (dialog->kbd_btn),
        (const char *) (get_kbd_name (code)));
}

inline const gchar *
get_kbd_value ()
{
    return (const gchar *) get (G_OBJECT (dialog->kbd_btn));
}

inline static const gchar *
get_init_kbd_value ()
{
    return (const gchar *) g_object_get_data (G_OBJECT (dialog->dialog), "init_kbd");
}

void
restore_kbd_config ()
{
    const gchar *tmp = get_init_kbd_value ();
    apply_kbd_config (tmp);
}

/* timezone */
void
inline set_tz_user_defined (gboolean b)
{
    g_object_set_data (G_OBJECT (dialog->tz_btn), "user_defined",
        (gpointer) b);
}

gboolean
inline get_tz_user_defined ()
{
    return (gboolean) g_object_get_data (G_OBJECT (dialog->tz_btn), "user_defined");
}

void
set_tz (const gchar *code)
{
    set (G_OBJECT (dialog->tz_btn), (gpointer) g_strdup (code));
    gtk_button_set_label (GTK_BUTTON (dialog->tz_btn),
        code);
}

inline const gchar *
get_tz_value ()
{
    return (const gchar *) get (G_OBJECT (dialog->tz_btn));
}

gboolean
get_x_is_custom ()
{
    gchar *err, *out;
    gint ret;
    gboolean retval = FALSE;

    g_spawn_command_line_sync ("sh " LIBEXECDIR "/xorg-conf.sh", &out, &err, &ret, NULL);
    if (WEXITSTATUS (ret) == 0)
    {
        gchar **spl;

        retval = TRUE;
        spl = g_strsplit (out, " ", -1);
        dialog->xorg_config->width = atoi (spl[0]);
        dialog->xorg_config->height = atoi (spl[1]);
        dialog->xorg_config->refresh = atoi (spl[2]);

        g_strfreev (spl);
    }
    g_free (out);
    g_free (err);

    return retval;
}

inline void
set_xorg_auto_config (gboolean b)
{
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (dialog->auto_xorg_chk), b);
}

inline gboolean
get_xorg_auto_config (void)
{
    return gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dialog->auto_xorg_chk));
}

/* show next time */
inline void
set_next_time (gboolean nt)
{
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (dialog->next_time), nt);
}

inline gboolean
get_next_time ()
{
    return gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dialog->next_time));
}

/* no X */
inline void
set_no_x (gboolean nox)
{
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (dialog->nox_chk), nox);
}

inline const gchar *
get_no_x ()
{
    return gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dialog->nox_chk))
        ? " noX"
        : "";
}

/* no hwclock */
inline void
set_no_hwclock (gboolean nohwclock)
{
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (dialog->nohwclock_chk), nohwclock);
}

inline const gchar *
get_no_hwclock ()
{
    return gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dialog->nohwclock_chk))
        ? " nohwclock"
        : "";
}

/* checkfs */
inline void
set_checkfs (gboolean check)
{
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (dialog->checkfs_chk), check);
}

inline const gchar *
get_checkfs ()
{
    return gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dialog->checkfs_chk))
        ? " checkfs"
        : "";
}

/* reboot schedule */
void
set_reboot_sched (gboolean sched, guint hour, guint minute)
{
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (dialog->reboot_chk), sched);
    gtk_spin_button_set_value (GTK_SPIN_BUTTON (dialog->reboot_min_spin),
                               (gdouble) minute);
    gtk_spin_button_set_value (GTK_SPIN_BUTTON (dialog->reboot_hour_spin),
                               (gdouble) hour);
}

gboolean
get_reboot_sched (guint *hour, guint *minute)
{
    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dialog->reboot_chk)))
    {
        *minute = (guint) gtk_spin_button_get_value (
                            GTK_SPIN_BUTTON (dialog->reboot_min_spin));
        *hour = (guint) gtk_spin_button_get_value (
                            GTK_SPIN_BUTTON (dialog->reboot_hour_spin));
        return TRUE;
    }
    else
    {
        *minute = 0;
        *hour = 0;
        return FALSE;
    }
}

/* persistence */
inline void
set_persistent (const gchar *persistent)
{
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (dialog->lock_chk),
                                  (strncmp (persistent, "full", 4) != 0));
}

inline const gchar *
get_persistent ()
{
    return gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dialog->lock_chk))
        ? ""
        : " persistent=full";
}

inline const gchar *
get_persistent_value ()
{
    return gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dialog->lock_chk))
        ? ""
        : "full";
}

/* lock active*/
void
set_lock (gboolean locked)
{
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (dialog->lock_chk), locked);
}

inline gboolean
get_lock ()
{
    return gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dialog->lock_chk));
}

/* reboot */
inline void
set_reboot_needed (gboolean needed)
{
    g_object_set_data (G_OBJECT (dialog->dialog), "reboot",
        (gpointer) needed);
}

inline gboolean
get_reboot_needed ()
{
    return (gboolean) g_object_get_data (G_OBJECT (dialog->dialog), "reboot");
}

/* update kbd and tz if needed */
void
update_locale_settings (const gchar *locale)
{
    gboolean kbd_user = get_kbd_user_defined ();
    gboolean tz_user = get_tz_user_defined ();
    if (kbd_user && tz_user)
        return;

    gchar *kbd, *tz;
    get_from_language (locale, &kbd, &tz);

    if (!kbd_user && (kbd[0] != 0))
    {
        set_kbd ((const gchar *)kbd);
        apply_kbd_config ((const gchar *)kbd);
    }
    if (!tz_user && (tz[0] != 0))
        set_tz ((const gchar *)tz);

    g_free (kbd);
    g_free (tz);
}


static gpointer
get_initial_conf (int id)
{
    switch (id)
    {
        case INIT_LOCALE:
            return g_object_get_data (G_OBJECT (dialog->dialog),
                "init_locale");
        case INIT_PERSISTENT:
            return g_object_get_data (G_OBJECT (dialog->dialog),
                "init_persistent");
        case INIT_LOCK:
            return g_object_get_data (G_OBJECT (dialog->dialog),
                "init_lock");
    }
    return (gpointer) NULL;
}

static void
set_initial_conf (gchar *persistent, gboolean lock,
                  gchar *kbd, gchar *locale)
{
    g_object_set_data (G_OBJECT (dialog->dialog), "init_locale",
        (gpointer) g_strdup (locale));
    g_object_set_data (G_OBJECT (dialog->dialog), "init_persistent",
        (gpointer) g_strdup (persistent));
    g_object_set_data (G_OBJECT (dialog->dialog), "init_kbd",
        (gpointer) g_strdup (kbd));
    g_object_set_data (G_OBJECT (dialog->dialog), "init_lock",
        (gpointer) lock);
}

static gboolean
need_reboot ()
{
    const gchar *init, *now;
    gboolean linit, lnow;

    /* locale */
    init = (const gchar *) get_initial_conf (INIT_LOCALE);
    now = get_locale_value ();
    if (strcmp (init, now) != 0)
        return TRUE;

    /* persistence */
    init = (const gchar *) get_initial_conf (INIT_PERSISTENT);
    now = get_persistent_value ();
    if (strcmp (init, now) != 0)
        return TRUE;

    /* lock */
    linit = (gboolean) get_initial_conf (INIT_LOCK);
    lnow = get_lock ();
    if (linit != lnow)
        return TRUE;

    return FALSE;
}

/* apply settings */
static void
change_next_time ()
{
    gchar *cmd;

    if (get_next_time ())
        cmd = g_strdup_printf ("sh " LIBEXECDIR "/apply-startup.sh true");
    else
        cmd = g_strdup_printf ("sh " LIBEXECDIR "/apply-startup.sh false");

    _system (cmd);
    g_free (cmd);
}

static void
schedule_reboot ()
{
    gchar *cmd = NULL;
    guint reboot_min;
    guint reboot_hour;

    if (get_reboot_sched (&reboot_hour, &reboot_min))
        cmd = g_strdup_printf ("sh " LIBEXECDIR "/reboot.sh %d %d",
                               reboot_hour, reboot_min);
    else
        cmd = g_strdup_printf ("sh " LIBEXECDIR "/reboot.sh cancel");

    _system (cmd);
    g_free (cmd);
}

static void
read_locale (gchar **locale)
{
    gchar *tmp = NULL, *p;
    FILE *f;
    gchar buf[128];

    f = fopen ("/etc/environment", "r");
    if (!f)
        goto out;

    while (fgets (buf, 128, f) != NULL)
    {
        if (strncmp (buf, "LANG=", 5) == 0)
        {
            p = strstr (buf, "\n");
            if (p)
                *p = '\0';
            tmp = g_shell_unquote (buf+5, NULL);
            break;
        }
    }

    fclose (f);

out:
    if (!tmp)
        tmp = g_strdup ("en_US.UTF-8");

    *locale = tmp;
}

static void
write_locale (gchar *locale)
{
    gchar *cmd;

    cmd = g_strdup_printf ("sh " LIBEXECDIR "/setlocale.sh %s", locale);
    system (cmd);
    g_free (cmd);
}

static void
read_keyboard (gchar **keyboard)
{
    gchar *tmp_layout = NULL, *tmp_variant = NULL, *tmp;
    gchar *p;
    FILE *f;
    gchar buf[128];

    f = fopen ("/etc/default/keyboard", "r");
    if (!f)
        goto out;

    while (fgets (buf, 128, f) != NULL)
    {
        if (strncmp (buf, "XKBLAYOUT=", 10) == 0)
        {
            p = strstr (buf, "\n");
            if (p)
                *p = '\0';
            tmp_layout = g_shell_unquote (buf+10, NULL);
        }
        else if (strncmp (buf, "XKBVARIANT=", 11) == 0)
        {
            p = strstr (buf, "\n");
            if (p)
                *p = '\0';
            tmp_variant = g_shell_unquote (buf+11, NULL);
            if (tmp_variant[0] == '\0')
            {
                g_free (tmp_variant);
                tmp_variant = NULL;
            }
        }
    }

    fclose (f);

out:
    if (!tmp_layout)
        tmp_layout = g_strdup ("us");

    if (tmp_variant)
        tmp = g_strdup_printf ("%s_%s", tmp_layout, tmp_variant);
    else
        tmp = g_strdup (tmp_layout);

    g_free (tmp_layout);
    g_free (tmp_variant);

    *keyboard = tmp;
}

static void
write_keyboard (gchar *keyboard)
{
    gchar *cmd;

    cmd = g_strdup_printf ("sh " LIBEXECDIR "/setkbd.sh %s", keyboard);
    system (cmd);
    g_free (cmd);
}

static void
read_timezone (gchar **timezone)
{
    gchar *tmp = NULL, *p;
    FILE *f;
    gchar buf[128];

    f = fopen ("/etc/timezone", "r");
    if (!f)
        goto out;

    if (!fgets (buf, 128, f))
    {
        fclose (f);
        goto out;
    }

    p = strstr (buf, "\n");
    if (p)
        *p = '\0';
    tmp = g_strdup (buf);
    fclose (f);

out:
    if (!tmp)
        tmp = g_strdup ("Europe/London");

    *timezone = tmp;
}

static void
write_timezone (gchar *timezone)
{
    gchar *cmd;

    cmd = g_strdup_printf ("sh " LIBEXECDIR "/settz.sh %s", timezone);
    system (cmd);
    g_free (cmd);
}

/**********************************/
void
read_config ()
{
    FILE *fp;
    gchar nexttime_txt[64];
    gchar *tmp = NULL;

    gchar *tz_txt = NULL;
    gchar *locale_txt = NULL;
    gchar *kbd_code_txt = NULL;
    gchar *screen_txt = g_strdup ("");
    gchar *persistent = g_strdup ("");
    guint reboot_min = 0;
    guint reboot_hour = 0;
    gboolean nox = FALSE;
    gboolean nohwclock = FALSE;
    gboolean checkfs = FALSE;
    gboolean nexttime = TRUE;
    gboolean lock = FALSE;
    gboolean custom_x;

    read_timezone (&tz_txt);
    read_locale (&locale_txt);
    read_keyboard (&kbd_code_txt);

    dialog->base_cmdline = g_strdup ("append ");

    dialog->lock_pwd = NULL;

    if (G_LIKELY((fp = fopen(CONFFILE, "r")))) {
        gchar *gline = NULL, *lline = NULL;
        gchar *tmp;
        size_t len = 0;
        ssize_t read;

        while ((read = getline(&tmp, &len, fp)) != -1) {
            if (!gline && (strncmp(tmp, "append  ", 8) == 0)) {
                g_free (gline);
                gline = g_strdup(tmp);
                if (gline[strlen(gline) - 1] == '\n')
                    gline[strlen(gline) - 1] = '\0';
            }
            if (!lline
                && (strncmp(tmp, "MENU MASTER PASSWD", 18) == 0)) {
                    lock = TRUE;
                    lline = g_strdup(tmp);
                    if (lline[strlen(lline) - 1] == '\n')
                        lline[strlen(lline) - 1] = '\0';
                    gchar **exploded = g_strsplit(lline, " ", 0);
                    if (g_strv_length(exploded) == 4)
                        dialog->lock_pwd =
                            g_strdup(exploded[3]);
                    g_strfreev(exploded);
                }
        }
        g_free(lline);
        g_free(tmp);
        fclose(fp);

        if (G_LIKELY(gline)) {
            gchar **exploded = g_strsplit(gline, " ", 0);
            int count = 0;
            while (exploded[count] != '\0') {
                if (*exploded[count] == '\0')
                {
                    count++;
                    continue;
                }
                gchar **item =
                    g_strsplit(exploded[count], "=", 0);
                if (g_strv_length(item) == 2) {
                    if (item[0] && item[1]
                        && strncmp(item[0], "persistent", 10) == 0)
                    {
                        g_free (persistent);
                        persistent = g_strdup (item[1]);
                    }
                    /* we don't handle the item, store it */
                    else
                    {
                        gchar *tmp = g_strconcat (dialog->base_cmdline, " ", exploded[count], NULL);
                        g_free (dialog->base_cmdline);
                        dialog->base_cmdline = tmp;
                    }

                } else {
                    if (strncmp (exploded[count], "noX", 3) == 0)
                        nox = TRUE;
                    else if (strncmp (exploded[count], "nohwclock", 9) == 0)
                        nohwclock = TRUE;
                    else if (strncmp (exploded[count], "checkfs", 7) == 0)
                        checkfs = TRUE;
                    else if ((strncmp (exploded[count], "append", 6) == 0)
                             || (strncmp (exploded[count], "lock", 4) == 0))
                    {
                        /* do nothing, but don't register it as an unknown parameter */
                    }
                    else
                    {
                        gchar *tmp = g_strconcat (dialog->base_cmdline, " ", exploded[count], NULL);
                        g_free (dialog->base_cmdline);
                        dialog->base_cmdline = tmp;
                    }
                }
                g_strfreev(item);
                count++;
            }
            g_free(gline);
            g_strfreev(exploded);
        }
    }

    g_sprintf(nexttime_txt, "%s/.config/autostart/lconfeditor.desktop",
          get_home());
    if (g_file_get_contents (nexttime_txt, &tmp, 0, NULL)
        && g_regex_match_simple ("X-GNOME-Autostart-enabled=false",
                                 tmp, G_REGEX_MULTILINE, 0))
    {
        nexttime = FALSE;
        g_free (tmp);
    }

    custom_x = get_x_is_custom ();

    if (reboot_scheduled (&reboot_hour, &reboot_min))
        set_reboot_sched (TRUE, reboot_hour, reboot_min);
    else
        set_reboot_sched (FALSE, reboot_hour, reboot_min);
    set_locale (locale_txt);
    set_kbd (kbd_code_txt);
    set_tz (tz_txt);
    set_next_time (nexttime);
    //set_xorg_auto_config (!custom_x);
    set_no_x (nox);
    set_no_hwclock (nohwclock);
    set_checkfs (checkfs);
    set_persistent (persistent);
    set_lock (lock);
    set_initial_conf (persistent, lock, kbd_code_txt, locale_txt);
    set_reboot_needed (FALSE);
    set_kbd_user_defined (FALSE);
    set_tz_user_defined (FALSE);

    g_free (tz_txt);
    g_free (locale_txt);
    g_free (kbd_code_txt);
    g_free (screen_txt);
    g_free (persistent);
}

void
write_config ()
{
    gchar *tz_txt = g_strdup (get_tz_value ());
    gchar *locale_txt = g_strdup (get_locale_value ());
    gchar *kbd_code_txt = g_strdup (get_kbd_value ());
    gchar *persistent = g_strdup (get_persistent ());
    gchar *nox = g_strdup (get_no_x ());
    gchar *nohwclock = g_strdup (get_no_hwclock ());
    gchar *checkfs = g_strdup (get_checkfs ());
    gboolean lock = get_lock ();
    gchar *str = NULL;
    gchar output[BUFSIZ];
    FILE *f;

    write_locale (locale_txt);
    write_timezone (tz_txt);
    write_keyboard (kbd_code_txt);

    str = g_strconcat (
        dialog->base_cmdline,
        persistent,
        nox,
        nohwclock,
        checkfs,
        NULL);

    if (lock)
        g_sprintf (output, locked_syscfg, dialog->lock_pwd, str);
    else
        g_sprintf (output, default_syscfg, str);
    if ((f = fopen (CONFFILE, "w"))) {
        fputs (output, f);
        fclose (f);
    }

    /*
    if (!get_xorg_auto_config ())
    {
        gchar *cmdline;

        cmdline = g_strdup_printf ("sh " LIBEXECDIR "/xorg-conf.sh set %d %d %d",
                                   dialog->xorg_config->width,
                                   dialog->xorg_config->height,
                                   dialog->xorg_config->refresh);
        _system (cmdline);
        g_free (cmdline);
    }
    else
    {
        g_rename ("/cdrom/etc/xorg.conf", "/cdrom/etc/xorg.conf.bak");
    }
    */

    /* blacklist lightdm service if noX */
    if (!strcmp (nox, ""))
        _system ("services-blacklist -n del lightdm");
    else
        _system ("services-blacklist -n add lightdm");

    /* apply other settings */
    change_next_time ();
    schedule_reboot ();
    if (need_reboot ())
        set_reboot_needed (TRUE);

    g_free (tz_txt);
    g_free (locale_txt);
    g_free (kbd_code_txt);
    g_free (persistent);
    g_free (nox);
    g_free (checkfs);
    g_free (str);
}
